/*=============================================================================
OPENP2X spinning gears
  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2006
  -- License restrictions apply (GPL v2)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
  Boston, MA  02110-1301  USA

  -- For details, see the file named "LICENSE.GPL2"
=============================================================================*/
/*
        My first attempt to create a GL application.
        It shows what I had in mind when designing
        the first logo of OpenP2X.
*/

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <math.h>
#include <sched.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#define RAD2DEG(a) ((double)(a) * 180 / M_PI)
#define DEG2RAD(a) ((double)(a) * M_PI / 180)
#define SLASH(v) (floor(v * 1000) / 1000)

// Options
#define DIST_A 2.0
#define DIST_B 2.3
#define DIST_C 0.1
#define MF 0.8 /* multfact for color boundaries */
#define NODES 16
#define GEARS 3

// Definitions
typedef struct rgb_s {
    GLubyte r, g, b;
} rgb_t;

typedef struct scheme_s {
    rgb_t color[4];
} scheme_t;

typedef struct {
    double heading, attitude, bank, speed, rade;
    const scheme_t *scheme;
} gear_t;

// Functions
static void init_gl(int, int);
static void start_timer(void);
static void sigalrm(int);
static void rotate(double, double, double, double *, double *,
                    double *, double *);
void resize_glscene(int, int);
static void draw_object(int, gear_t *, double, double);
static void draw_glscene(void);
static void key_pressed(unsigned char, int, int);
static inline double rand_d(double);

// Variables
static gear_t Gear[GEARS];
static int window;
static int paused = 0;
static long frames = 0;
static time_t start;

//-----------------------------------------------------------------------------
int main(int argc, const char **argv) {
    glutInit(&argc, (char **)argv);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_ALPHA | GLUT_DEPTH);
    glutInitWindowSize(640, 480);
    glutInitWindowPosition(0, 0);
    window = glutCreateWindow("OpenP2X rotating O");
    glutDisplayFunc(draw_glscene);
    //glutFullScreen();
    glutIdleFunc(draw_glscene);
    glutReshapeFunc(resize_glscene);
    glutKeyboardFunc(key_pressed);
    init_gl(640, 480);
    glutMainLoop();
    return 1;
}

//-----------------------------------------------------------------------------
static void init_gl(int width, int height) {
    static const scheme_t Scheme[] = {
        {{  {255, 255, 255}, {220, 220, 220},
            {255, 209, 165}, {255, 169, 85}}},
        {{  {255, 255, 255}, {220, 220, 220},
            {220, 232, 188}, {189, 211, 129}}},
        {{  {255, 255, 255}, {228, 228, 228},
            {216, 236, 255}, {141, 200, 255}}},
    };
    int gi;

    glClearColor(255, 255, 255, 0);
    glClearDepth(1);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glShadeModel(GL_SMOOTH);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
    glEnable(GL_LINE_SMOOTH);
    glLineWidth(0.5);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45, (double)width / height, 0.1, 100);
    glMatrixMode(GL_MODELVIEW);

    srand(time(NULL) ^ getpid());
    for(gi = 0; gi < GEARS; ++gi) {
        gear_t *gp = &Gear[gi];
        gp->heading  = rand_d(360);
        gp->attitude = rand_d(360);
        gp->bank     = rand_d(360);
        gp->speed    = 3; // rand() % 10;
        gp->rade     = gi * 0.7;
        gp->scheme   = &Scheme[gi];
    }

    start_timer();
    return;
}

static void start_timer(void) {
    struct sigaction sa;

    sa.sa_handler = sigalrm;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = SA_RESTART;

    sigaction(SIGALRM, &sa, NULL);
    start = time(NULL);
    alarm(1);
    return;
}

static void sigalrm(int s) {
    time_t stop = time(NULL);
    time_t delta = stop - start;
    printf("%ld frames in %ld seconds = %3.3f fps\n",
     frames, (long)delta, (double)frames / delta);
    alarm(1);
    return;
}

static void rotate(double heading, double attitude, double bank,
 double *ap, double *xp, double *yp, double *zp)
{
    // Assuming the angles are in radians.
    double a, x, y, z, n,
     c1 = cos(heading/2), c2 = cos(attitude/2), c3 = cos(bank/2),
     s1 = sin(heading/2), s2 = sin(attitude/2), s3 = sin(bank/2),
     C1C2 = c1 * c2, S1S2 = s1 * s2,
     C1S2 = c1 * s2, S1C2 = s1 * c2;

    a = 2 * acos(C1C2 * c3 - S1S2 * s3);
    x = C1C2 * s3 + S1S2 * c3;
    y = S1C2 * c3 + C1S2 * s3;
    z = C1S2 * c3 - S1C2 * s3;

    if((n = x * x + y * y + z * z) < 0.00001) {
        x = z = 0;
        y = 1;
    } else {
        n = sqrt(n);
        x /= n;
        y /= n;
        z /= n;
    }
    *ap = RAD2DEG(SLASH(a));
    *xp = SLASH(x);
    *yp = SLASH(y);
    *zp = SLASH(z);
    return;
}

void resize_glscene(int width, int height) {
    if(height == 0)
        height = 1;
    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45, (double)width / height, 0.1, 100);
    glMatrixMode(GL_MODELVIEW);
    return;
}

/* The main drawing function. */

static void draw_object(int FRAMEMODE, gear_t *gp, double s, double e) {
    double da = DIST_A + gp->rade, db = DIST_B + gp->rade;
    glBegin(FRAMEMODE); {
        // front
        glVertex3d(da * sin(s), da * cos(s), 0 + DIST_C);
        glVertex3d(db * sin(s), db * cos(s), 0 + DIST_C);
        glVertex3d(db * sin(e), db * cos(e), 0 + DIST_C);
        glVertex3d(da * sin(e), da * cos(e), 0 + DIST_C);
    } glEnd();
    glBegin(FRAMEMODE); {
        // back
        glVertex3d(da * sin(s), da * cos(s), 0 - DIST_C);
        glVertex3d(db * sin(s), db * cos(s), 0 - DIST_C);
        glVertex3d(db * sin(e), db * cos(e), 0 - DIST_C);
        glVertex3d(da * sin(e), da * cos(e), 0 - DIST_C);
    } glEnd();
    glBegin(FRAMEMODE); {
        // left
        glVertex3d(da * sin(s), da * cos(s), 0 + DIST_C);
        glVertex3d(da * sin(s), da * cos(s), 0 - DIST_C);
        glVertex3d(db * sin(s), db * cos(s), 0 - DIST_C);
        glVertex3d(db * sin(s), db * cos(s), 0 + DIST_C);
    } glEnd();
    glBegin(FRAMEMODE); {
        // right
        glVertex3d(da * sin(e), da * cos(e), 0 + DIST_C);
        glVertex3d(da * sin(e), da * cos(e), 0 - DIST_C);
        glVertex3d(db * sin(e), db * cos(e), 0 - DIST_C);
        glVertex3d(db * sin(e), db * cos(e), 0 + DIST_C);
    } glEnd();
    glBegin(FRAMEMODE); {
        // top
        glVertex3d(db * sin(s), db * cos(s), 0 + DIST_C);
        glVertex3d(db * sin(s), db * cos(s), 0 - DIST_C);
        glVertex3d(db * sin(e), db * cos(e), 0 - DIST_C);
        glVertex3d(db * sin(e), db * cos(e), 0 + DIST_C);
    } glEnd();
    glBegin(FRAMEMODE); {
        // bottom
        glVertex3d(da * sin(s), da * cos(s), 0 + DIST_C);
        glVertex3d(da * sin(s), da * cos(s), 0 - DIST_C);
        glVertex3d(da * sin(e), da * cos(e), 0 - DIST_C);
        glVertex3d(da * sin(e), da * cos(e), 0 + DIST_C);
    } glEnd();
    return;
}

static void draw_glscene(void) {
    int gi;

    if(paused) {
        usleep(100000);
        return;
    }

    glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
    ++frames;

    for(gi = 0; gi < GEARS; ++gi) {
        double ra, rx, ry, rz;
        gear_t *gp = &Gear[gi];
        int n;

        glLoadIdentity();
        glTranslated(0, 0, -10);

        gp->bank     += .1  * gp->speed;
        gp->attitude -= .01 * gp->speed;

        rotate(DEG2RAD(gp->heading), DEG2RAD(gp->attitude), DEG2RAD(gp->bank),
          &ra, &rx, &ry, &rz);
        glRotated(ra, rx, ry, rz);

        for(n = 0; n < NODES; ++n) {
            double s = n * 2 * M_PI / NODES + M_PI / (2 * NODES);
            double e = (n + 1) * 2 * M_PI / NODES;
            const rgb_t *rgb;
            GLubyte r, g, b;
            int idx;

            if(n == NODES -1 || n == 1) idx = 2;
            else if(n == 0)             idx = 3;
            else                        idx = 1;

            rgb = &gp->scheme->color[idx];
            r   = rgb->r;
            g   = rgb->g;
            b   = rgb->b;

            glColor3ub(r, g, b);
            draw_object(GL_POLYGON, gp, s, e);
            glColor3ub(r *= MF, g *= MF, b *= MF);
            draw_object(GL_LINE_LOOP, gp, s, e);
        }
    }

    // swap the buffers to display, since double buffering is used.
    glutSwapBuffers();
    return;
}

static void key_pressed(unsigned char key, int x, int y) {
    if(key == '\e') {
        glutDestroyWindow(window);
        exit(0);
    } else if(key == ' ') {
        paused = !paused;
    }
    return;
}

static inline double rand_d(double m) {
    return (double)rand() * m / RAND_MAX;
}

//=============================================================================
